@@ -1,6 +1,6 @@
 package Bio::Graphics::FeatureFile;
 
-# $Id$
+# $Id: FeatureFile.pm,v 1.60 2004/11/18 03:11:32 scain Exp $
 # This package parses and renders a simple tab-delimited format for features.
 # It is simpler than GFF, but still has a lot of expressive power.
 # See __END__ for the file format
@@ -53,7 +53,7 @@ FeatureFile object has been initialized, you can interrogate it for
 its consistuent features and their settings, or render the entire file
 onto a Bio::Graphics::Panel.
 
-This moduel is a precursor of Jason Stajich's
+This module is a precursor of Jason Stajich's
 Bio::Annotation::Collection class, and fulfills a similar function of
 storing a collection of sequence features.  However, it also stores
 rendering information about the features, and does not currently
@@ -116,14 +116,17 @@ belong to the same group named yk53c10.
 
 use strict;
 use Bio::Graphics::Feature;
-use Bio::Graphics::Panel;
+use Bio::DB::GFF::Util::Rearrange;
 use Carp;
+use Bio::DB::GFF;
 use IO::File;
 use Text::Shellwords;
 
 # default colors for unconfigured features
 my @COLORS = qw(cyan blue red yellow green wheat turquoise orange);
+
 use constant WIDTH => 600;
+use constant MAX_REMAP => 100;
 
 =head2 METHODS
 
@@ -190,7 +193,6 @@ sub new {
   my $self = bless {
 		    config   => {},
 		    features => {},
-		    groups   => {},
 		    seenit   => {},
 		    types    => [],
 		    max      => undef,
@@ -201,7 +203,7 @@ sub new {
 		   },$class;
   $self->{coordinate_mapper} = $args{-map_coords} 
     if exists $args{-map_coords} && ref($args{-map_coords}) eq 'CODE';
-  $self->{smart_features}    = $args{-smart_features} if exists $args{-smart_features};
+  $self->smart_features($args{-smart_features})       if exists $args{-smart_features};
   $self->{safe}              = $args{-safe}           if exists $args{-safe};
 
   # call with
@@ -258,7 +260,7 @@ sub render {
   my $color;
   my %types = map {$_=>1} $self->configured_types;
 
-  my @configured_types   = grep {exists $self->features->{$_}} $self->configured_types;
+  my @configured_types   = grep {exists $self->{features}{$_}} $self->configured_types;
   my @unconfigured_types = sort grep {!exists $types{$_}}      $self->types;
 
   my @base_config = $self->style('general');
@@ -282,7 +284,11 @@ sub render {
   }
 
   for my $type (@configured_types,@unconfigured_types) {
-    my $features = $self->features($type);
+    my $f = $self->features($type);
+    my @features = grep {$self->{visible}{$_} || $_->type eq 'group'} @$f;
+    next unless @features;  # suppress tracks for features that don't appear
+    my $features = \@features;
+
     my @auto_bump;
     push @auto_bump,(-bump  => @$features < $max_bump)  if defined $max_bump;
     push @auto_bump,(-label => @$features < $max_label) if defined $max_label;
@@ -367,7 +373,7 @@ sub parse_file {
   $self->init_parse;
   while (<$fh>) {
     chomp;
-    $self->parse_line($_);
+    $self->parse_line($_) || last;
   }
   $self->finish_parse;
 }
@@ -388,8 +394,20 @@ sub parse_line {
   local $_ = shift;
 
   s/\015//g;  # get rid of carriage returns left over by MS-DOS/Windows systems
+  s/\s+$//;   # get rid of trailing whitespace
 
-  return if /^\s*[\#]/;
+  # capture GFF header
+  if (/^\#\#gff-version\s+(\d+)/) {
+    $self->{gff_version} = $1;
+    require Bio::DB::GFF;
+    return 1;
+  }
+
+  # skip on blank lines and comments
+  return 1 if /^\s*[\#]/;
+
+  # abort if we see a >FASTA line
+  return 0 if /^>/;
 
   if (/^\s+(.+)/ && $self->{current_tag}) { # continuation line
     my $value = $1;
@@ -398,7 +416,7 @@ sub parse_line {
     # respect newlines in code subs
     $self->{config}{$cc}{$self->{current_tag}} .= "\n"
       if $self->{config}{$cc}{$self->{current_tag}}=~ /^sub\s*\{/;
-    return;
+    return 1;
   }
 
   if (/^\s*\[([^\]]+)\]/) {  # beginning of a configuration section
@@ -406,7 +424,7 @@ sub parse_line {
     my $cc = $label =~ /^(general|default)$/i ? 'general' : $label;  # normalize
     push @{$self->{types}},$cc unless $cc eq 'general';
     $self->{current_config} = $cc;
-    return;
+    return 1;
   }
 
   if (/^([\w: -]+?)\s*=\s*(.*)/) {   # key value pair within a configuration section
@@ -415,45 +433,62 @@ sub parse_line {
     my $value = defined $2 ? $2 : '';
     $self->{config}{$cc}{$tag} = $value;
     $self->{current_tag} = $tag;
-    return;
+    return 1;
   }
 
 
   if (/^$/) { # empty line
     undef $self->{current_tag};
-    return;
+    return 1;
   }
 
   # parse data lines
-  my @tokens = eval { shellwords($_||'') };
+  my @tokens = shellwords($_);
+  # my @tokens = split "\t",$_;  # faster?
   unshift @tokens,'' if /^\s+/;
 
   # close any open group
-  undef $self->{grouptype} if length $tokens[0] > 0;
-
-  if (@tokens < 3) {      # short line; assume a group identifier
-    $self->{grouptype}     = shift @tokens;
-    $self->{groupname}     = shift @tokens;
-    return;
+  if (length $tokens[0] > 0 && $self->{group}) {
+    push @{$self->{features}{$self->{grouptype}}},$self->{group};
+    undef $self->{group};
+    undef $self->{grouptype};
   }
 
-  my($ref,$type,$name,$strand,$bounds,$description,$url);
+  if (@tokens < 3) {      # short line; assume a group identifier
+    my $type               = shift @tokens;
+    my $name               = shift @tokens;
+    $self->{group}         = Bio::Graphics::Feature->new(-name => $name,
+							 -type => 'group');
+    $self->{grouptype}     = $type;
+    return 1;
+  }
+
+  my($ref,$type,$name,$strand,$bounds,$description,$url,$score,%attributes);
+
+  my @parts;
 
   if (@tokens >= 8) { # conventional GFF file
-    my ($r,$source,$method,$start,$stop,$score,$s,$phase,@rest) = @tokens;
+    my ($r,$source,$method,$start,$stop,$scor,$s,$phase,@rest) = @tokens;
     my $group = join ' ',@rest;
-    $type   = join(':',$method,$source);
-    $bounds = join '..',$start,$stop;
+    $type   = defined $source && $source ne '.' ? join(':',$method,$source) : $method;
+    #$bounds = join '..',$start,$stop;
+    @parts   = ([$start,$stop]);
     $strand = $s;
     if ($group) {
       my ($notes,@notes);
-      (undef,$self->{groupname},undef,undef,$notes) = split_group($group);
+      (undef,$name,undef,undef,$notes) = $self->split_group($group);
       foreach (@$notes) {
-	if (m!^(http|ftp)://!) { $url = $_ } else { push @notes,$_ }
+	my ($key,$value) = @$_;
+	if ($value =~ m!^(http|ftp)://!) { 
+	  $url = $_ 
+	} elsif ($key=~/note/i) { 
+	  push @notes,$value;
+	}
       }
       $description = join '; ',@notes if @notes;
+      $score       = $scor if defined $scor && $scor ne '.';
     }
-    $name ||= $self->{groupname};
+    $name ||= $self->{group}->display_id if $self->{group};
     $ref = $r;
   }
 
@@ -476,24 +511,36 @@ sub parse_line {
   }
   $self->{refs}{$ref}++ if defined $ref;
 
-  my @parts = map { [/(-?\d+)(?:-|\.\.)(-?\d+)/]} split /(?:,| )\s*/,$bounds;
+  @parts = map { [/(-?\d+)(?:-|\.\.)(-?\d+)/]} split /(?:,| )\s*/,$bounds
+    if $bounds && !@parts;
 
   foreach (@parts) { # max and min calculation, sigh...
     $self->{min} = $_->[0] if !defined $self->{min} || $_->[0] < $self->{min};
     $self->{max} = $_->[1] if !defined $self->{max} || $_->[1] > $self->{max};
   }
 
+  my $visible = 1;
   if ($self->{coordinate_mapper} && $ref) {
-    ($ref,@parts) = $self->{coordinate_mapper}->($ref,@parts);
-    return unless $ref;
+    my @remapped = $self->{coordinate_mapper}->($ref,@parts);
+    ($ref,@parts) = @remapped if @remapped;
+    $visible   = @remapped;
+    return 1 if !$visible && $self->{feature_count} > MAX_REMAP;
   }
 
   $type = '' unless defined $type;
   $name = '' unless defined $name;
 
+  # if strand is not explicitly given in file, we infer it
+  # from the order of start and end coordinates
+  # (this is to deal with confusing documentation, actually)
+  unless (defined $strand) {
+    foreach (@parts) {
+      $strand           ||= $_->[0] <= $_->[1] ? '+' : '-';
+      ($_->[0],$_->[1])   = ($_->[1],$_->[0]) if $_->[0] > $_->[1];
+    }
+  }
+
   # attribute handling
-  my %attributes;
-  my $score;
   if (defined $description && $description =~ /\w+=\w+/) { # attribute line
     my @attributes = split /;\s*/,$description;
     foreach (@attributes) {
@@ -516,26 +563,34 @@ sub parse_line {
 
   # either create a new feature or add a segment to it
   if (my $feature = $self->{seenit}{$type,$name}) {
-    $feature->add_segment(@parts);
-  } else {
-    $feature = $self->{seenit}{$type,$name} = 
-      Bio::Graphics::Feature->new(-name       => $name,
-				  -type       => $type,
-				  $strand ? (-strand   => make_strand($strand)) : (),
-				  defined $score ? (-score=>$score) : (),
-				  -segments   => \@parts,
-				  -desc       => $description,
-				  -ref        => $ref,
-				  -attributes => \%attributes,
-				  defined($url) ? (-url      => $url) : (),
-				 );
+
+    # create a new segment to hold the parts
+    if (!$feature->segments) {
+      my $new_segment  = bless {%$feature},ref $feature;
+      $feature->add_segment($new_segment);
+    }
+    # add the segments
+    $feature->add_segment(map {
+      _make_feature($name,$type,$strand,$description,$ref,\%attributes,$url,$score,[$_])
+    }  @parts);
+    $self->{visible}{$feature}++  if $visible;
+  }
+
+  else {
+    $feature = $self->{seenit}{$type,$name} = _make_feature($name,$type,$strand,
+							    $description,$ref,
+							    \%attributes,$url,$score,\@parts);
     $feature->configurator($self) if $self->smart_features;
-    if ($self->{grouptype}) {
-      push @{$self->{groups}{$self->{grouptype}}{$self->{groupname}}},$feature;
+    if ($self->{group}) {
+      $self->{group}->add_segment($feature);
     } else {
       push @{$self->{features}{$type}},$feature;  # for speed; should use add_feature() instead
+      $self->{visible}{$feature}++  if $visible;
+      $self->{feature_count}++;
     }
   }
+
+  return 1;
 }
 
 sub _unescape {
@@ -546,6 +601,22 @@ sub _unescape {
   @_;
 }
 
+sub _make_feature {
+  my ($name,$type,$strand,$description,$ref,$attributes,$url,$score,$parts) = @_;
+  my @coordinates = @$parts > 1 ? (-segments => $parts) : (-start=>$parts->[0][0],-end=>$parts->[0][1]);
+  Bio::Graphics::Feature->new(-name       => $name,
+			      -type       => $type,
+			      -subtype    => "${type}_part",
+			      $strand ? (-strand   => make_strand($strand)) : (),
+			      -desc       => $description,
+			      -ref        => $ref,
+			      -attributes => $attributes,
+			      defined $url   ? (-url  => $url) : (),
+			      defined $score ? (-score=>$score) : (),
+			      @coordinates,
+			     );
+}
+
 =over 4
 
 =item $features-E<gt>add_feature($feature [=E<gt>$type])
@@ -563,7 +634,10 @@ feature's primary_tag() method will be invoked to get the type.
 sub add_feature {
   my $self = shift;
   my ($feature,$type) = @_;
+  $feature->configurator($self) if $self->smart_features;
   $type = $feature->primary_tag unless defined $type;
+  $self->{visible}{$feature}++;
+  $self->{feature_count}++;
   push @{$self->{features}{$type}},$feature;
 }
 
@@ -630,12 +704,12 @@ sub set {
 }
 
 # break circular references
-sub destroy {
+sub finished {
   my $self = shift;
   delete $self->{features};
 }
 
-sub DESTROY { shift->destroy(@_) }
+sub DESTROY { shift->finished(@_) }
 
 =over 4
 
@@ -662,6 +736,9 @@ Call with no elements to retrieve all stanza names:
 
 sub setting {
   my $self = shift;
+  if (@_ > 2) {
+    $self->{config}->{$_[0]}{$_[1]} = $_[2];
+  }
   if ($self->safe) {
      $self->code_setting(@_);
   } else {
@@ -672,11 +749,17 @@ sub setting {
 # return configuration information
 # arguments are ($type) => returns tags for type
 #               ($type=>$tag) => returns values of tag on type
+#               ($type=>$tag,$value) => sets value of tag
 sub _setting {
   my $self = shift;
   my $config = $self->{config} or return;
   return keys %{$config} unless @_;
   return keys %{$config->{$_[0]}} if @_ == 1;
+
+  # this seems reasonable, but it prevents deliberate assignment of
+  # undef to settings
+  # return $config->{$_[0]}{$_[1]}  if @_ > 1 && exists $config->{$_[0]};
+
   return $config->{$_[0]}{$_[1]}  if @_ > 1;
 }
 
@@ -711,7 +794,7 @@ sub code_setting {
     $self->set($section,$option,$coderef);
     return $coderef;
   }
-  elsif ($setting =~ /^sub\s*\{/) {
+  elsif ($setting =~ /^sub\s*(\(\$\$\))*\s*\{/) {
     my $coderef   = eval $setting;
     warn $@ if $@;
     $self->set($section,$option,$coderef);
@@ -762,7 +845,11 @@ sub style {
   my $type = shift;
 
   my $config  = $self->{config}  or return;
-  my $hashref = $config->{$type} or return;
+  my $hashref = $config->{$type};
+  unless ($hashref) {
+    $type =~ s/:.+$//;
+    $hashref = $config->{$type} or return;
+  }
 
   return map {("-$_" => $hashref->{$_})} keys %$hashref;
 }
@@ -826,6 +913,56 @@ sub types {
   return keys %{$features};
 }
 
+=over 4
+
+=item $features = $features-E<gt>features($type)
+
+Return a list of all the feature types of type "$type".  If the
+featurefile object was created by parsing a file or text scalar, then
+the features will be of type Bio::Graphics::Feature (which follow the
+Bio::FeatureI interface).  Otherwise the list will contain objects of
+whatever type you added with calls to add_feature().
+
+Two APIs:
+
+  1) original API:
+
+      # Reference to an array of all features of type "$type"
+      $features = $features-E<gt>features($type)
+
+      # Reference to an array of all features of all types
+      $features = $features-E<gt>features()
+
+      # A list when called in a list context
+      @features = $features-E<gt>features()
+
+   2) Bio::Das::SegmentI API:
+
+       @features = $features-E<gt>features(-type=>['list','of','types']);
+
+       # variants
+       $features = $features-E<gt>features(-type=>['list','of','types']);
+       $features = $features-E<gt>features(-type=>'a type');
+       $iterator = $features-E<gt>features(-type=>'a type',-iterator=>1);
+
+=back
+
+=cut
+
+# return features
+sub features {
+  my $self = shift;
+  my ($types,$iterator,@rest) = defined($_[0] && $_[0]=~/^-/)
+    ? rearrange([['TYPE','TYPES']],@_) : (\@_);
+  $types = [$types] if $types && !ref($types);
+  my @types = ($types && @$types) ? @$types : $self->types;
+  my @features = map {@{$self->{features}{$_}}} @types;
+  if ($iterator) {
+    require Bio::Graphics::FeatureFile::Iterator;
+    return Bio::Graphics::FeatureFile::Iterator->new(\@features);
+  }
+  return wantarray ? @features : \@features;
+}
 
 =over 4
 
@@ -841,15 +978,6 @@ whatever type you added with calls to add_feature().
 
 =cut
 
-# return features
-sub features {
-  my $self = shift;
-  return $self->{features}{shift()} if @_;
-  return $self->{features};
-}
-
-
-
 sub make_strand {
   local $^W = 0;
   return +1 if $_[0] =~ /^\+/ || $_[0] > 0;
@@ -857,6 +985,151 @@ sub make_strand {
   return 0;
 }
 
+=head2 get_seq_stream
+
+ Title   : get_seq_stream
+ Usage   : $stream = $s->get_seq_stream(@args)
+ Function: get a stream of features that overlap this segment
+ Returns : a Bio::SeqIO::Stream-compliant stream
+ Args    : see below
+ Status  : Public
+
+This is the same as feature_stream(), and is provided for Bioperl
+compatibility.  Use like this:
+
+ $stream = $s->get_seq_stream('exon');
+ while (my $exon = $stream->next_seq) {
+    print $exon->start,"\n";
+ }
+
+=cut
+
+sub get_seq_stream {
+  my $self = shift;
+  local $^W = 0;
+  my @args = $_[0] =~ /^-/ ? (@_,-iterator=>1) : (-types=>\@_,-iterator=>1);
+  $self->features(@args);
+}
+
+=head2 get_feature_by_name
+
+ Usage   : $db->get_feature_by_name(-name => $name)
+ Function: fetch features by their name
+ Returns : a list of Bio::DB::GFF::Feature objects
+ Args    : the name of the desired feature
+ Status  : public
+
+This method can be used to fetch a named feature from the file.
+
+The full syntax is as follows.  Features can be filtered by
+their reference, start and end positions
+
+  @f = $db->get_feature_by_name(-name  => $name,
+                                -ref   => $sequence_name,
+                                -start => $start,
+                                -end   => $end);
+
+This method may return zero, one, or several Bio::Graphics::Feature
+objects.
+
+=cut
+
+sub get_feature_by_name {
+   my $self = shift;
+   my ($name,$ref,$start,$end) = rearrange(['NAME','REF','START','END'],@_);
+   my $match = <<'END';
+sub {
+        my $f = shift;
+END
+   if (defined $name) {
+      if ($name =~ /[\?\*]/) {  # regexp
+        $name =  quotemeta($name);
+        $name =~ s/\\\?/.?/g;
+        $name =~ s/\\\*/.*/g;
+        $match .= "     return unless \$f->display_name =~ /$name/i;\n";
+      } else {
+        $match .= "     return unless \$f->display_name eq '$name';\n";
+      }
+   }
+
+   if (defined $ref) {
+      $match .= "     return unless \$f->ref eq '$ref';\n";
+   }
+   if (defined $start && $start =~ /^-?\d+$/) {
+      $match .= "     return unless \$f->stop >= $start;\n";
+   }
+   if (defined $end && $end =~ /^-?\d+$/) {
+      $match .= "     return unless \$f->start <= $end;\n";
+   }
+   $match .= "     return 1;\n}";
+
+   my $match_sub = eval $match;
+   unless ($match_sub) {
+     warn $@;
+     return;
+   }
+
+   return grep {$match_sub->($_)} $self->features;
+}
+
+=head2 search_notes
+
+ Title   : search_notes
+ Usage   : @search_results = $db->search_notes("full text search string",$limit)
+ Function: Search the notes for a text string
+ Returns : array of results
+ Args    : full text search string, and an optional row limit
+ Status  : public
+
+Each row of the returned array is a arrayref containing the following fields:
+
+  column 1     Display name of the feature
+  column 2     The text of the note
+  column 3     A relevance score.
+
+=cut
+
+sub search_notes {
+  my $self = shift;
+  my ($search_string,$limit) = @_;
+  my @results;
+  my $search = join '|',map {quotemeta($_)} $search_string =~ /(\S+)/g;
+
+  for my $feature ($self->features) {
+    next unless $feature->{attributes};
+    my @attributes = $feature->all_tags;
+    my @values     = map {$feature->each_tag_value} @attributes;
+    push @values,$feature->notes        if $feature->notes;
+    push @values,$feature->display_name if $feature->display_name;
+    next unless @values;
+    my $value      = "@values";
+    my $matches    = 0;
+    my $note;
+    my @hits = $value =~ /($search)/ig;
+    $note ||= $value if @hits;
+    $matches += @hits;
+    next unless $matches;
+
+    my $relevance = 10 * $matches;
+    push @results,[$feature,$note,$relevance];
+    last if @results >= $limit;
+  }
+
+  @results;
+}
+
+
+=head2 get_feature_stream(), top_SeqFeatures(), all_SeqFeatures()
+
+Provided for compatibility with older BioPerl and/or Bio::DB::GFF
+APIs.
+
+=cut
+
+*get_feature_stream = \&get_seq_stream;
+*top_SeqFeatures    = *all_SeqFeatures = \&features;
+
+
 =over 4
 
 =item @refs = $features-E<gt>refs
@@ -901,29 +1174,19 @@ sub init_parse {
   my $s = shift;
 
   $s->{seenit} = {}; 
-  $s->{max}      = $s->{min} = undef;
-  $s->{types}    = [];
-  $s->{groups}   = {};
-  $s->{features} = {};
-  $s->{config}   = {}
+  $s->{max}         = $s->{min} = undef;
+  $s->{types}       = [];
+  $s->{features}    = {};
+  $s->{config}      = {};
+  $s->{gff_version} = 0;
+  $s->{feature_count}=0; 
 }
 
 sub finish_parse {
   my $s = shift;
-  $s->consolidate_groups;
   $s->evaluate_coderefs if $s->safe;
   $s->{seenit} = {};
-  $s->{groups} = {};
-}
-
-sub consolidate_groups {
-  my $self = shift;
-  my $groups = $self->{groups} or return;
-
-  for my $type (keys %$groups) {
-    my @groups = values %{$groups->{$type}};
-    push @{$self->{features}{$type}},@groups;
-  }
+  delete $s->{gff_version};
 }
 
 sub evaluate_coderefs {
@@ -953,54 +1216,16 @@ sub base2package {
 }
 
 sub split_group {
-  my $group = shift;
-
-  $group =~ s/\\;/$;/g;  # protect embedded semicolons in the group
-  $group =~ s/( \"[^\"]*);([^\"]*\")/$1$;$2/g;
-  my @groups = split(/\s*;\s*/,$group);
-  foreach (@groups) { s/$;/;/g }
-
-  my ($gclass,$gname,$tstart,$tstop,@notes);
-
-  foreach (@groups) {
-
-    my ($tag,$value) = /^(\S+)\s*(.*)/;
-    $value =~ s/\\t/\t/g;
-    $value =~ s/\\r/\r/g;
-    $value =~ s/^"//;
-    $value =~ s/"$//;
-
-    # if the tag is "Note", then we add this to the
-    # notes array
-   if ($tag eq 'Note') {  # just a note, not a group!
-     push @notes,$value;
-   }
-
-    # if the tag eq 'Target' then the class name is embedded in the ID
-    # (the GFF format is obviously screwed up here)
-    elsif ($tag eq 'Target' && $value =~ /([^:\"]+):([^\"]+)/) {
-      ($gclass,$gname) = ($1,$2);
-      ($tstart,$tstop) = /(\d+) (\d+)/;
-    }
-
-    elsif (!$value) {
-      push @notes,$tag;  # e.g. "Confirmed_by_EST"
-    }
-
-    # otherwise, the tag and value correspond to the
-    # group class and name
-    else {
-      ($gclass,$gname) = ($tag,$value);
-    }
-  }
-
-  return ($gclass,$gname,$tstart,$tstop,\@notes);
+  my $self = shift;
+  return Bio::DB::GFF->split_group(shift, $self->{gff_version} > 2);
 }
 
 # create a panel if needed
 sub new_panel {
   my $self = shift;
 
+  require Bio::Graphics::Panel;
+
   # general configuration of the image here
   my $width         = $self->setting(general => 'pixels')
                       || $self->setting(general => 'width')
@@ -1078,7 +1303,7 @@ sub feature2label {
 
 =over 4
 
-=item $link = $features-E<gt>make_link($feature)
+=item $link = $features-E<gt>link_pattern($linkrule,$feature,$panel)
 
 Given a feature, tries to generate a URL to link out from it.  This
 uses the 'link' option, if one is present.  This method is a
@@ -1088,41 +1313,76 @@ convenience for the generic genome browser.
 
 =cut
 
-sub make_link {
-  my $self     = shift;
-  my $feature  = shift;
-  for my $label ($self->feature2label($feature)) {
-    my $link     = $self->setting($label,'link');
-    $link        = $self->setting(general=>'link') unless defined $link;
-    next unless $link;
-    return $self->link_pattern($link,$feature);
-  }
-  return;
-}
-
 sub link_pattern {
-  my $self = shift;
-  my ($pattern,$feature,$panel) = @_;
+  my $self     = shift;
+  my ($linkrule,$feature,$panel) = @_;
+
+  $panel ||= 'Bio::Graphics::Panel';
+
+  if (ref($linkrule) && ref($linkrule) eq 'CODE') {
+    my $val = eval {$linkrule->($feature,$panel)};
+    warn $@ if $@;
+    return $val;
+  }
+
   require CGI unless defined &CGI::escape;
   my $n;
-  $pattern =~ s/\$(\w+)/
+  $linkrule =~ s/\$(\w+)/
     CGI::escape(
-    $1 eq 'ref'              ? ($n = $feature->location->seq_id) && "$n"
-      : $1 eq 'name'         ? ($n = $feature->display_name) && "$n"  # workaround broken CGI.pm
+    $1 eq 'ref'              ? (($n = $feature->location->seq_id) && "$n") || ''
+      : $1 eq 'name'         ? (($n = $feature->display_name) && "$n")     || ''
       : $1 eq 'class'        ? eval {$feature->class}  || ''
-      : $1 eq 'type'         ? eval {$feature->method} || $feature->primary_tag
-      : $1 eq 'method'       ? eval {$feature->method} || $feature->primary_tag
-      : $1 eq 'source'       ? eval {$feature->source} || $feature->source_tag
-      : $1 eq 'start'        ? $feature->start
-      : $1 eq 'end'          ? $feature->end
-      : $1 eq 'stop'         ? $feature->end
-      : $1 eq 'segstart'     ? $panel->start
-      : $1 eq 'segend'       ? $panel->end
+      : $1 eq 'type'         ? eval {$feature->method} || $feature->primary_tag || ''
+      : $1 eq 'method'       ? eval {$feature->method} || $feature->primary_tag || ''
+      : $1 eq 'source'       ? eval {$feature->source} || $feature->source_tag  || ''
+      : $1 eq 'start'        ? $feature->start || ''
+      : $1 eq 'end'          ? $feature->end   || ''
+      : $1 eq 'stop'         ? $feature->end   || ''
+      : $1 eq 'segstart'     ? $panel->start   || ''
+      : $1 eq 'segend'       ? $panel->end     || ''
       : $1 eq 'description'  ? eval {join '',$feature->notes} || ''
+      : $1 eq 'id'           ? $feature->feature_id || ''
       : $1
-	       )
-       /exg;
-  return $pattern;
+       )
+	/exg;
+  return $linkrule;
+}
+
+sub make_link {
+  my $self             = shift;
+  my ($feature,$panel) = @_;
+
+  for my $label ($self->feature2label($feature)) {
+    my $linkrule     = $self->setting($label,'link');
+    $linkrule        = $self->setting(general=>'link') unless defined $linkrule;
+    return $self->link_pattern($linkrule,$feature,$panel);
+  }
+}
+
+sub make_title {
+  my $self = shift;
+  my $feature = shift;
+  my $method  = $feature->method;
+  my $seqid   = $feature->can('seq_id')      ? $feature->seq_id : $feature->location->seq_id;
+  my $title = eval {
+    if ($feature->can('target') && (my $target = $feature->target)) {
+      join (' ',
+	    $method,
+	    (defined $seqid ? "$seqid:" : '').
+	    $feature->start."..".$feature->end,
+	    $feature->target.':'.
+	    $feature->target->start."..".$feature->target->end);
+    } else {
+      join(' ',
+	   $method,
+	   $feature->can('display_name') ? $feature->display_name : $feature->info,
+	   (defined $seqid ? "$seqid:" : '').
+	   ($feature->start||'?')."..".($feature->end||'?')
+	  );
+    }
+  };
+  warn $@ if $@;
+  $title;
 }
 
 # given a feature type, return its label(s)
