@@ -1,4 +1,4 @@
-# $Id$
+# $Id: GenericHit.pm,v 1.34 2004/12/29 21:02:13 jason Exp $
 #
 # BioPerl module for Bio::Search::Hit::GenericHit
 #
@@ -19,19 +19,30 @@ Bio::Search::Hit::GenericHit - A generic implementation of the Bio::Search::Hit:
     use Bio::Search::Hit::GenericHit;
     my $hit = new Bio::Search::Hit::GenericHit(-algorithm => 'blastp');
 
-    # more likely
+    # typically one gets HitI objects from a SearchIO stream via a ResultI
     use Bio::SearchIO;
     my $parser = new Bio::SearchIO(-format => 'blast', -file => 'result.bls');
 
     my $result = $parser->next_result;
     my $hit    = $result->next_hit;
 
+# TODO: Describe how to configure a SearchIO stream so that it generates
+#       GenericHit objects.
 
 =head1 DESCRIPTION
 
 This object handles the hit data from a Database Sequence Search such
 as FASTA or BLAST.
 
+Unless you're writing a parser, you won't ever need to create a
+GenericHit or any other HitI-implementing object. If you use
+the SearchIO system, HitI objects are created automatically from
+a SearchIO stream which returns Bio::Search::Hit::HitI objects.
+
+For documentation on what you can do with GenericHit (and other HitI
+objects), please see the API documentation in
+L<Bio::Search::Hit::HitI|Bio::Search::Hit::HitI>.
+
 =head1 FEEDBACK
 
 =head2 Mailing Lists
@@ -46,20 +57,15 @@ the Bioperl mailing list.  Your participation is much appreciated.
 =head2 Reporting Bugs
 
 Report bugs to the Bioperl bug tracking system to help us keep track
-of the bugs and their resolution. Bug reports can be submitted via
-email or the web:
+of the bugs and their resolution. Bug reports can be submitted via the
+web:
 
-  bioperl-bugs@bioperl.org
   http://bugzilla.bioperl.org/
 
 =head1 AUTHOR - Jason Stajich and Steve Chervitz
 
-Email jason@bioperl.org
-Email sac@bioperl.org
-
-=head1 CONTRIBUTORS
-
-Additional contributors names and emails here
+Email jason-at-bioperl-dot-org
+Email sac-at-bioperl-dot-org
 
 =head1 APPENDIX
 
@@ -93,10 +99,12 @@ require Bio::Search::SearchUtils;
            -accession    => Accession number (optional)
            -length       => Length of the Hit (optional)
            -score        => Raw Score for the Hit (optional)
+           -bits         => Bit Score for the Hit (optional)
            -significance => Significance value for the Hit (optional)
            -algorithm    => Algorithm used (BLASTP, FASTX, etc...)
            -hsps         => Array ref of HSPs for this Hit. 
-           -iteration    => integer for the PSI-Blast iteration number
+           -found_again  => boolean, true if hit appears in a 
+                            "previously found" section of a PSI-Blast report.
 
 =cut
 
@@ -106,17 +114,17 @@ sub new {
   my $self = $class->SUPER::new(@args);
   my ($hsps, $name,$query_len,$desc, $acc, $locus, $length,
       $score,$algo,$signif,$bits,
-      $iter,$rank) = $self->_rearrange([qw(HSPS 
-				     NAME 
-				     QUERY_LEN
-				     DESCRIPTION
-				     ACCESSION
-				     LOCUS
-				     LENGTH SCORE ALGORITHM 
-				     SIGNIFICANCE BITS ITERATION
-				     RANK )], @args);
+      $rank) = $self->_rearrange([qw(HSPS 
+                                     NAME 
+                                     QUERY_LEN
+                                     DESCRIPTION
+                                     ACCESSION
+                                     LOCUS
+                                     LENGTH SCORE ALGORITHM 
+                                     SIGNIFICANCE BITS
+                                     RANK )], @args);
   
-  $self->{'_query_length'} = $query_len;
+  defined $query_len && $self->query_length($query_len);
 
   if( ! defined $name ) { 
       $self->throw("Must have defined a valid name for Hit");
@@ -132,20 +140,25 @@ sub new {
   defined $signif && $self->significance($signif);
   defined $score  && $self->raw_score($score);
   defined $bits   && $self->bits($bits);
-  defined $iter   && $self->iteration($iter);
   defined $rank   && $self->rank($rank);
 
   $self->{'_iterator'} = 0;
-  $self->{'_hsps'} = [];
   if( defined $hsps  ) {
       if( ref($hsps) !~ /array/i ) {
-	  $self->warn("Did not specify a valid array ref for the param HSPS ($hsps)");
+          $self->warn("Did not specify a valid array ref for the param HSPS ($hsps)");
       } else {
-	  while( @$hsps ) { 
-	      $self->add_hsp(shift @$hsps );
-	  }
+          my $hspcount=0;
+          while( @{$hsps} ) { 
+              $hspcount++;
+              $self->add_hsp(shift @{$hsps} );
+          }
+          $self->{'_hsps'} = undef if $hspcount == 0;
       }
   } 
+  else {
+      $self->{'_hsps'} = undef;
+  }
+
   return $self;
 }
 
@@ -163,9 +176,10 @@ sub new {
 sub add_hsp {
    my ($self,$hsp) = @_;
    if( !defined $hsp || ! $hsp->isa('Bio::Search::HSP::HSPI') ) { 
-       $self->warn("Must provide a valid Bio::Search::HSP::HSPI object to object: $self method: add_hsp value: $hsp");
+       $self->throw("Must provide a valid Bio::Search::HSP::HSPI object to object: $self method: add_hsp value: $hsp");
        return undef;
    }
+#   print STDERR "GenericHit::add_hsp()\n";
    push @{$self->{'_hsps'}}, $hsp;
    return scalar @{$self->{'_hsps'}};
 }
@@ -190,8 +204,8 @@ sub name {
     my ($self,$value) = @_;
     my $previous = $self->{'_name'};
     if( defined $value || ! defined $previous ) {
-	$value = $previous = '' unless defined $value;
-	$self->{'_name'} = $value;
+        $value = $previous = '' unless defined $value;
+        $self->{'_name'} = $value;
     } 
     return $previous;
 }
@@ -210,10 +224,10 @@ sub accession {
     my ($self,$value) = @_;
     my $previous = $self->{'_accession'};
     if( defined $value || ! defined $previous ) { 
-	$value = $previous = '' unless defined $value;
-	$self->{'_accession'} = $value;
+        $value = $previous = '' unless defined $value;
+        $self->{'_accession'} = $value;
     } 
-	return $previous;
+        return $previous;
 }
 
 =head2 description
@@ -230,8 +244,8 @@ sub description {
     my ($self,$value) = @_;
     my $previous = $self->{'_description'};
     if( defined $value || ! defined $previous ) { 
-	$value = $previous = '' unless defined $value;
-	$self->{'_description'} = $value;
+        $value = $previous = '' unless defined $value;
+        $self->{'_description'} = $value;
     } 
     return $previous;
 }
@@ -250,8 +264,8 @@ sub length {
     my ($self,$value) = @_;
     my $previous = $self->{'_length'};
     if( defined $value || ! defined $previous ) { 
-	$value = $previous = 0 unless defined $value;
-	$self->{'_length'} = $value;
+        $value = $previous = 0 unless defined $value;
+        $self->{'_length'} = $value;
     } 
     return $previous;
 }
@@ -275,8 +289,8 @@ sub algorithm {
     my ($self,$value) = @_;
     my $previous = $self->{'_algorithm'};
     if( defined $value || ! defined $previous ) { 
-	$value = $previous = '' unless defined $value;
-	$self->{'_algorithm'} = $value;
+        $value = $previous = '' unless defined $value;
+        $self->{'_algorithm'} = $value;
     } 
     return $previous;
 }
@@ -297,12 +311,20 @@ sub raw_score {
     my ($self,$value) = @_;
     my $previous = $self->{'_score'};
     if( defined $value || ! defined $previous ) { 
-	$value = $previous = '' unless defined $value;
-	$self->{'_score'} = $value;
+        $value = $previous = '' unless defined $value;
+        $self->{'_score'} = $value;
     } 
     return $previous;
 }
 
+=head2 score
+
+Equivalent to L<raw_score()|raw_score>
+
+=cut
+
+sub score { shift->raw_score(@_); }
+
 =head2 significance
 
  Title   : significance
@@ -319,10 +341,17 @@ sub raw_score {
 sub significance {
     my ($self,$value) = @_;
     my $previous = $self->{'_significance'};
-    if( defined $value || ! defined $previous ) { 
-	$value = $previous = '' unless defined $value;
-	$self->{'_significance'} = $value;
-    } 
+    if( defined $value ) { 
+        $self->{'_significance'} = $value;
+    } elsif ( ! defined $previous ) {
+	unless( defined $self->{'_hsps'}->[0] ) {
+	    $self->warn("No HSPs for this Hit (".$self->name.")");
+	    return undef;
+	}
+        # Set the significance of the Hit to that of the top HSP.
+        $previous = $self->{'_significance'} = ($self->hsps)[0]->significance;
+    }
+
     return $previous;
 }
 
@@ -342,14 +371,19 @@ See Also   : L<score()|score>
 #---------
 sub bits { 
 #---------
-    my ($self) = @_; 
-    
-    my $bits = $self->{'_bits'};
-    if( ! defined $bits ) {
-	$bits = $self->{'_hsps'}->[0]->bits();
-	$self->{'_bits'} = $bits;
-    } 
-    return $bits;
+    my ($self,$value) = @_; 
+    my $previous = $self->{'_bits'};
+    if( defined $value ) { 
+        $self->{'_bits'} = $value;
+    } elsif ( ! defined $previous ) {
+        # Set the significance of the Hit to that of the top HSP.
+	unless( defined $self->{'_hsps'}->[0] ) {
+	    $self->warn("No HSPs for this Hit (".$self->name.")");
+	    return undef;
+	}
+        $previous = $self->{'_bits'} = ($self->hsps)[0]->bits;
+    }    
+    return $previous;
 }
 
 =head2 next_hsp
@@ -364,9 +398,11 @@ sub bits {
 =cut
 
 sub next_hsp {
-    my ($self) = @_;
+    my $self = shift;
     $self->{'_iterator'} = 0 unless defined $self->{'_iterator'};
-    return undef if $self->{'_iterator'} > scalar @{$self->{'_hsps'}};
+    return undef unless
+        defined($self->{'_hsps'}) 
+        && $self->{'_iterator'} <= scalar @{$self->{'_hsps'}};
     return $self->{'_hsps'}->[$self->{'_iterator'}++];    
 }
 
@@ -393,15 +429,11 @@ sub hsps {
 #---------
    my $self = shift;
    
-   if (not ref $self->{'_hsps'}) {
-       $self->throw("Can't get HSPs: data not collected.");
-   }
-   
    return wantarray 
        #  returning list containing all HSPs.
-       ? @{$self->{'_hsps'}}
+       ? @{$self->{'_hsps'} || []}
    #  returning number of HSPs.
-   : scalar(@{$self->{'_hsps'}});
+   : scalar(@{$self->{'_hsps'} || []});
 }
 
 =head2 num_hsps
@@ -421,10 +453,15 @@ See Also   : L<hsps()|hsps>
 sub num_hsps {
     my $self = shift;
     
-    if (not defined $self->{'_hsps'}) {
-	$self->throw("Can't get HSPs: data not collected.");
+    unless ($self->{'_hsps'}) {
+        #return wantarray ? ('-','-') : '-';
+        return '-';
     }
 
+#    if (not defined $self->{'_hsps'}) {
+#        $self->throw("Can't get HSPs: data not collected.");
+#    }
+
     return scalar(@{$self->{'_hsps'}});
 }
 
@@ -432,7 +469,7 @@ sub num_hsps {
 
  Title   : rewind
  Usage   : $hit->rewind;
- Function: Allow one to reset the HSP iteration to the beginning
+ Function: Allow one to reset the HSP iterator to the beginning
            Since this is an in-memory implementation
  Returns : none
  Args    : none
@@ -444,27 +481,6 @@ sub rewind{
    $self->{'_iterator'} = 0;
 }
 
-=head2 iteration
-
- Title   : iteration
- Usage   : $obj->iteration($newval)
- Function: PSI-BLAST iteration
- Returns : value of iteration
- Args    : newvalue (optional)
-
-
-=cut
-
-sub iteration{
-   my ($self,$value) = @_;
-   if( defined $value) {
-      $self->{'_psiblast_iteration'} = $value;
-    }
-    return $self->{'_psiblast_iteration'};
-
-}
-
-
 =head2 ambiguous_aln
 
  Usage     : $ambig_code = $hit_object->ambiguous_aln();
@@ -482,7 +498,7 @@ sub iteration{
            :          of similar domains.
  Argument  : n/a
  Throws    : n/a
- Status    : Experimental
+ Comment   : Note: "sbjct" is synonymous with "hit"
 
 =cut
 
@@ -496,7 +512,7 @@ sub ambiguous_aln {
 
 =head2 overlap
 
-See documentation in L<Bio::Search::SearchUtils::overlap()|Bio::Search::SearchUtils>
+See documentation in L<Bio::Search::Hit::HitI::overlap()|Bio::Search::Hit::HitI>
 
 =cut
 
@@ -542,9 +558,11 @@ sub n {
 
     my ($n);
     if(not defined($self->{'_n'})) {
-	$n = $self->hsp->n;
+	if( $self->hsp ) {
+	    $n = $self->hsp->n;
+	}
     } else {
-	$n = $self->{'_n'}; 
+        $n = $self->{'_n'}; 
     } 
     $n ||= $self->num_hsps;
 
@@ -629,7 +647,7 @@ sub hsp {
     $option ||= 'best';
     
     if (not ref $self->{'_hsps'}) {
-	$self->throw("Can't get HSPs: data not collected.");
+        $self->throw("Can't get HSPs: data not collected.");
     }
 
     my @hsps = @{$self->{'_hsps'}};
@@ -638,7 +656,7 @@ sub hsp {
     return $hsps[$#hsps] if $option =~ /worst|last/i;
 
     $self->throw("Can't get HSP for: $option\n" .
-		 "Valid arguments: 'best', 'worst'");
+                 "Valid arguments: 'best', 'worst'");
 }
 
 =head2 logical_length
@@ -646,15 +664,23 @@ sub hsp {
  Usage     : $hit_object->logical_length( [seq_type] );
            : (mostly intended for internal use).
  Purpose   : Get the logical length of the hit sequence.
-           : If the Blast is a TBLASTN or TBLASTX, the returned length 
-           : is the length of the would-be amino acid sequence (length/3).
-           : For all other BLAST flavors, this function is the same as length().
+           : This is necessary since the number of identical/conserved residues 
+           : can be in terms of peptide sequence space, yet the query and/or hit
+           : sequence are in nucleotide space.
  Example   : $len    = $hit_object->logical_length();
  Returns   : Integer 
  Argument  : seq_type = 'query' or 'hit' or 'sbjct' (default = 'query')
              ('sbjct' is synonymous with 'hit')
  Throws    : n/a
- Comments  : This is important for functions like frac_aligned_query()
+ Comments  :
+           : In the case of BLAST flavors:
+           : For TBLASTN reports, the length of the aligned portion of the 
+           : nucleotide hit sequence is divided by 3; for BLASTX reports, 
+           : the length of the aligned portion of the nucleotide query 
+           : sequence is divided by 3. For TBLASTX reports, the length of 
+           : both hit and query sequence are converted.
+           :
+           : This is important for functions like frac_aligned_query()
            : which need to operate in amino acid coordinate space when dealing
            : with [T]BLAST[NX] type reports.
 
@@ -669,23 +695,21 @@ sub logical_length {
     my $seqType = shift || 'query';
     $seqType = 'sbjct' if $seqType eq 'hit';
 
-    my $length;
+    my ($length, $logical);
+    my $algo = $self->algorithm;
 
     # For the sbjct, return logical sbjct length
     if( $seqType eq 'sbjct' ) {
-	$length = $self->{'_logical_length'} || $self->{'_length'};
-    }
-    else {
+        $length = $self->length;
+    } else {
         # Otherwise, return logical query length
-        $length = $self->{'_query_length'};
-	$self->throw("Must have defined query_len") unless ( $length );
-
-        # Adjust length based on BLAST flavor.
-        if($self->algorithm =~ /T?BLASTX/ ) {
-            $length /= 3;
-        }
+        $length = $self->query_length();
+        $self->throw("Must have defined query_len") unless ( $length );
     }
-    return $length;
+
+    $logical = Bio::Search::SearchUtils::logical_length($algo, $seqType, $length);
+
+    return int($logical);
 }
 
 =head2 length_aln
@@ -716,23 +740,30 @@ See Also   : L<length()|length>, L<frac_aligned_query()|frac_aligned_query>, L<f
 sub length_aln {
 #---------------
     my( $self, $seqType, $num ) = @_;
-    
+
     $seqType ||= 'query';
     $seqType = 'sbjct' if $seqType eq 'hit';
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
-
+    # Setter:
     if( defined $num) {
         return $self->{'_length_aln_'.$seqType} = $num;
     }
 
+    unless ($self->{'_hsps'}) {
+        #return wantarray ? ('-','-') : '-';
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        return '-';
+    }
+
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
     my $data = $self->{'_length_aln_'.$seqType};
     
     ## If we don't have data, figure out what went wrong.
     if(!$data) {
-	$self->throw("Can't get length aln for sequence type \"$seqType\". " . 
-		     "Valid types are 'query', 'hit', 'sbjct' ('sbjct' = 'hit')");
-    }		
+        $self->throw("Can't get length aln for sequence type \"$seqType\". " . 
+                     "Valid types are 'query', 'hit', 'sbjct' ('sbjct' = 'hit')");
+    }                
     return $data;
 }    
 
@@ -778,7 +809,13 @@ sub gaps {
     $seqType ||= (wantarray ? 'list' : 'total');
     $seqType = 'sbjct' if $seqType eq 'hit';
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        return wantarray ? ('-','-') : '-';
+        #return '-';
+    }
+
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
 
     $seqType = lc($seqType);
 
@@ -788,12 +825,12 @@ sub gaps {
         return $self->{'_gaps_'.$seqType} = $num;
     }
     elsif($seqType =~ /list|array/i) {
-	return ($self->{'_gaps_query'}, $self->{'_gaps_sbjct'});
+        return ($self->{'_gaps_query'}, $self->{'_gaps_sbjct'});
     }
     elsif($seqType eq 'total') {
-	return ($self->{'_gaps_query'} + $self->{'_gaps_sbjct'}) || 0;
+        return ($self->{'_gaps_query'} + $self->{'_gaps_sbjct'}) || 0;
     } else {
-	return $self->{'_gaps_'.$seqType} || 0;
+        return $self->{'_gaps_'.$seqType} || 0;
     }
 }    
 
@@ -810,30 +847,39 @@ sub matches {
     my( $self, $arg1, $arg2) = @_;
     my(@data,$data);
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        return wantarray ? ('-','-') : '-';
+    }
 
-    if(!$arg1) {
-	@data = ($self->{'_totalIdentical'}, $self->{'_totalConserved'});
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
 
-	return @data if @data;
+    unless( $arg1 ) {
+        @data = ($self->{'_totalIdentical'}, $self->{'_totalConserved'});
 
+        return @data;
     } else {
 
         if( defined $arg2 ) {
-	    $self->{'_totalIdentical'} = $arg1;
-	    $self->{'_totalConserved'} = $arg2;
+            $self->{'_totalIdentical'} = $arg1;
+            $self->{'_totalConserved'} = $arg2;
             return ( $arg1, $arg2 );
         }
-	elsif($arg1 =~ /^id/i) { 
-	    $data = $self->{'_totalIdentical'};
-	} else {
-	    $data = $self->{'_totalConserved'};
-	}
-	return $data if $data;
+        elsif($arg1 =~ /^id/i) { 
+            $data = $self->{'_totalIdentical'};
+        } else {
+            $data = $self->{'_totalConserved'};
+        }
+        #print STDERR "\nmatches(): id=$self->{'_totalIdentical'}, cons=$self->{'_totalConserved'}\n\n";
+        return $data;
     }
     
-    ## Something went wrong if we make it to here.
-    $self->throw("Can't get identical or conserved data: no data.");
+    ## If we make it to here, it is likely the case that
+    ## the parser constructed a minimal hit object from the summary line only.
+    ## It either delibrately skipped parsing the alignment section,
+    ## or was not able to because it was absent (due to blast executable parameter
+    ## setting such as -b 0 (B=0 for WU-BLAST) )
+    #$self->throw("Can't get identical or conserved data: no data.");
 }
 
 
@@ -868,6 +914,11 @@ sub start {
 #----------
     my ($self, $seqType, $num) = @_;
 
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        return wantarray ? ('-','-') : '-';
+    }
+
     $seqType ||= (wantarray ? 'list' : 'query');
     $seqType = 'sbjct' if $seqType eq 'hit';
 
@@ -880,14 +931,14 @@ sub start {
     if($self->num_hsps == 1) {
         return $self->hsp->start($seqType);
     } else {
-        Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
-        if($seqType =~ /list|array/i) {
-            return ($self->{'_queryStart'}, $self->{'_sbjctStart'});
-        } else {
-            ## Sensitive to member name changes.
-            $seqType = "_\L$seqType\E";
-            return $self->{$seqType.'Start'};
-        }
+        &Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+	if($seqType =~ /list|array/i) {
+	    return ($self->{'_queryStart'}, $self->{'_sbjctStart'});
+	} else {
+	    ## Sensitive to member name changes.
+	    $seqType = "_\L$seqType\E";
+	    return $self->{$seqType.'Start'};
+	}
     }
 }
 
@@ -896,19 +947,23 @@ sub start {
 
  Usage     : $sbjct->end( [seq_type] );
  Purpose   : Gets the end coordinate for the query, sbjct, or both sequences
-           : in the BlastHit object. If there is more than one HSP, the largest end
+           : in the BlastHit object. If there is more than one HSP, 
+             the largest end
            : value of all HSPs is returned.
  Example   : $qend = $sbjct->end('query');
            : $send = $sbjct->end('hit');
            : ($qend, $send) = $sbjct->end();
  Returns   : scalar context: integer
-           : array context without args: list of two integers (queryEnd, sbjctEnd)
-           : Array context can be "induced" by providing an argument of 'list' or 'array'.
+           : array context without args: list of two integers 
+           : (queryEnd, sbjctEnd)
+           : Array context can be "induced" by providing an argument 
+           : of 'list' or 'array'.
  Argument  : In scalar context: seq_type = 'query' or 'sbjct'
            :  (case insensitive). If not supplied, 'query' is used.
  Throws    : n/a
- Comments  : This method requires that all HSPs be tiled. If there is more than one
-           : HSP and they have not already been tiled, they will be tiled first automatically..
+ Comments  : This method requires that all HSPs be tiled. If there is 
+           : more than one HSP and they have not already been tiled, 
+           : they will be tiled first automatically..
            : Remember that the start and end coordinates of all HSPs are 
            : normalized so that start < end. Strand information can be
            : obtained by calling $hit->strand().
@@ -922,10 +977,13 @@ sub end {
 #----------
     my ($self, $seqType, $num) = @_;
 
+    unless ($self->{'_hsps'}) {
+        return wantarray ? ('-','-') : '-';
+    }
+
     $seqType ||= (wantarray ? 'list' : 'query');
     $seqType = 'sbjct' if $seqType eq 'hit';
 
-
     if( defined $num ) {
         $seqType = "_\L$seqType\E";
         return $self->{$seqType.'Stop'} = $num;
@@ -933,16 +991,16 @@ sub end {
 
     # If there is only one HSP, defer this call to the solitary HSP.
     if($self->num_hsps == 1) {
-	return $self->hsp->end($seqType);
+        return $self->hsp->end($seqType);
     } else {
-	Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
-	if($seqType =~ /list|array/i) {
-	    return ($self->{'_queryStop'}, $self->{'_sbjctStop'});
-	} else {
-	    ## Sensitive to member name changes.
-	    $seqType = "_\L$seqType\E";
-	    return $self->{$seqType.'Stop'};
-	}
+        Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+        if($seqType =~ /list|array/i) {
+            return ($self->{'_queryStop'}, $self->{'_sbjctStop'});
+        } else {
+            ## Sensitive to member name changes.
+            $seqType = "_\L$seqType\E";
+            return $self->{$seqType.'Stop'};
+        }
     }
 }
 
@@ -984,7 +1042,19 @@ sub range {
            : default = 'query' (but see comments below).
            : ('sbjct' is synonymous with 'hit')
  Throws    : n/a
- Comments  : Different versions of Blast report different values for the total
+ Comments  :
+           : To compute the fraction identical, the logical length of the 
+           : aligned portion of the sequence is used, meaning that
+           : in the case of BLAST flavors, for TBLASTN reports, the length of 
+           : the aligned portion of the 
+           : nucleotide hit sequence is divided by 3; for BLASTX reports, 
+           : the length of the aligned portion of the nucleotide query 
+           : sequence is divided by 3. For TBLASTX reports, the length of 
+           : both hit and query sequence are converted.
+           : This is necessary since the number of identical residues is
+           : in terms of peptide sequence space.
+           :
+           : Different versions of Blast report different values for the total
            : length of the alignment. This is the number reported in the
            : denominators in the stats section:
            : "Identical = 34/120 Positives = 67/120".
@@ -1019,10 +1089,16 @@ sub frac_identical {
     ## Sensitive to member name format.
     $seqType = lc($seqType);
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    my $ident = $self->{'_totalIdentical'};
-    my $total = $self->{'_length_aln_'.$seqType};
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    my $ident = $self->matches('id');
+    my $total = $self->length_aln($seqType);
     my $ratio = $ident / $total;
     my $ratio_rounded = sprintf( "%.3f", $ratio);
 
@@ -1032,7 +1108,6 @@ sub frac_identical {
 }
 
 
-
 =head2 frac_conserved
 
  Usage     : $hit_object->frac_conserved( [seq_type] );
@@ -1045,7 +1120,19 @@ sub frac_identical {
            : default = 'query' (but see comments below).
            : ('sbjct' is synonymous with 'hit')
  Throws    : n/a
- Comments  : Different versions of Blast report different values for the total
+ Comments  :
+           : To compute the fraction conserved, the logical length of the 
+           : aligned portion of the sequence is used, meaning that
+           : in the case of BLAST flavors, for TBLASTN reports, the length of 
+           : the aligned portion of the 
+           : nucleotide hit sequence is divided by 3; for BLASTX reports, 
+           : the length of the aligned portion of the nucleotide query 
+           : sequence is divided by 3. For TBLASTX reports, the length of 
+           : both hit and query sequence are converted.
+           : This is necessary since the number of conserved residues is
+           : in terms of peptide sequence space.
+           :
+           : Different versions of Blast report different values for the total
            : length of the alignment. This is the number reported in the
            : denominators in the stats section:
            : "Positives = 34/120 Positives = 67/120".
@@ -1080,10 +1167,16 @@ sub frac_conserved {
     ## Sensitive to member name format.
     $seqType = lc($seqType);
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    my $consv = $self->{'_totalConserved'};
-    my $total = $self->{'_length_aln_'.$seqType};
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    my $consv = $self->matches('cons');
+    my $total = $self->length_aln($seqType);
     my $ratio = $consv / $total;
     my $ratio_rounded = sprintf( "%.3f", $ratio);
 
@@ -1107,11 +1200,6 @@ sub frac_conserved {
  Throws    : n/a
  Comments  : If you need data for each HSP, use hsps() and then interate
            : through the HSP objects.
-           : To compute the fraction aligned, the logical length of the query
-           : sequence is used, meaning that for [T]BLASTX reports, the 
-           : full length of the query sequence is converted into amino acids
-           : by dividing by 3. This is necessary because of the way 
-           : the lengths of aligned sequences are computed.
            : This method requires that all HSPs be tiled. If they have not
            : already been tiled, they will be tiled first automatically.
 
@@ -1124,10 +1212,16 @@ sub frac_aligned_query {
 #----------------------
     my $self = shift;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    sprintf( "%.2f", $self->{'_length_aln_query'}/ 
-	     $self->logical_length('query'));
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    sprintf( "%.2f", $self->length_aln('query') /
+             $self->logical_length('query'));
 }
 
 
@@ -1144,11 +1238,6 @@ sub frac_aligned_query {
  Throws    : n/a
  Comments  : If you need data for each HSP, use hsps() and then interate
            : through the HSP objects.
-           : To compute the fraction aligned, the logical length of the sbjct
-           : sequence is used, meaning that for TBLAST[NX] reports, the 
-           : full length of the sbjct sequence is converted into amino acids
-           : by dividing by 3. This is necessary because of the way 
-           : the lengths of aligned sequences are computed.
            : This method requires that all HSPs be tiled. If they have not
            : already been tiled, they will be tiled first automatically.
 
@@ -1161,9 +1250,15 @@ sub frac_aligned_hit {
 #--------------------
     my $self = shift;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    sprintf( "%.2f", $self->{'_length_aln_sbjct'}/$self->logical_length('sbjct'));
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    sprintf( "%.2f", $self->length_aln('sbjct') / $self->logical_length('sbjct'));
 }
 
 
@@ -1175,9 +1270,7 @@ Same as L<frac_aligned_hit()|frac_aligned_hit>
 
 =cut
 
-#----------------
-sub frac_aligned_sbjct {  my $self=shift; $self->frac_aligned_hit(@_); }
-#----------------
+*frac_aligned_sbjct = \&fract_aligned_hit;
 
 =head2 num_unaligned_sbjct
 
@@ -1185,10 +1278,7 @@ Same as L<num_unaligned_hit()|num_unaligned_hit>
 
 =cut
 
-#----------------
-sub num_unaligned_sbjct {  my $self=shift; $self->num_unaligned_hit(@_); }
-#----------------
-
+*num_unaligned_sbjct = \&num_unaligned_hit;
 
 
 =head2 num_unaligned_hit
@@ -1216,9 +1306,15 @@ sub num_unaligned_hit {
 #---------------------
     my $self = shift;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    my $num = $self->logical_length('sbjct') - $self->{'_length_aln_sbjct'};
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    my $num = $self->logical_length('sbjct') - $self->length_aln('sbjct');
     ($num < 0 ? 0 : $num );
 }
 
@@ -1248,9 +1344,15 @@ sub num_unaligned_query {
 #-----------------------
     my $self = shift;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
 
-    my $num = $self->logical_length('query') - $self->{'_length_aln_query'};
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
+
+    my $num = $self->logical_length('query') - $self->length_aln('query');
     ($num < 0 ? 0 : $num );
 }
 
@@ -1294,14 +1396,14 @@ sub seq_inds {
 
     my (@inds, $hsp);
     foreach $hsp ($self->hsps) {
-	# This will merge data for all HSPs together.
-	push @inds, $hsp->seq_inds($seqType, $class);
+        # This will merge data for all HSPs together.
+        push @inds, $hsp->seq_inds($seqType, $class);
     }
     
     # Need to remove duplicates and sort the merged positions.
     if(@inds) {
-	my %tmp = map { $_, 1 } @inds;
-	@inds = sort {$a <=> $b} keys %tmp;
+        my %tmp = map { $_, 1 } @inds;
+        @inds = sort {$a <=> $b} keys %tmp;
     }
 
     $collapse ?  &Bio::Search::SearchUtils::collapse_nums(@inds) : @inds; 
@@ -1310,7 +1412,7 @@ sub seq_inds {
 
 =head2 strand
 
-See documentation in L<Bio::Search::SearchUtils::strand()|Bio::Search::SearchUtils>
+See documentation in L<Bio::Search::Hit::HitI::strand()|Bio::Search::Hit::HitI>
 
 =cut
 
@@ -1319,7 +1421,13 @@ sub strand {
 #----------
     my ($self, $seqType, $strnd) = @_;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        return wantarray ? ('-','-') : '-';
+        #return '-';
+    }
+
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
 
     $seqType ||= (wantarray ? 'list' : 'query');
     $seqType = 'sbjct' if $seqType eq 'hit';
@@ -1335,7 +1443,7 @@ sub strand {
     my ($qstr, $hstr);
     # If there is only one HSP, defer this call to the solitary HSP.
     if($self->num_hsps == 1) {
-	return $self->hsp->strand($seqType);
+        return $self->hsp->strand($seqType);
     } 
     elsif( defined $self->{'_strand_query'}) {
         # Get the data computed during hsp tiling.
@@ -1343,7 +1451,7 @@ sub strand {
         $hstr = $self->{'_strand_sbjct'}
     }
     else {
-	# otherwise, iterate through all HSPs collecting strand info.
+        # otherwise, iterate through all HSPs collecting strand info.
         # This will return the string "-1/1" if there are HSPs on different strands.
         # NOTE: This was the pre-10/21/02 procedure which will no longer be used,
         # (unless the above elsif{} is commented out).
@@ -1368,7 +1476,7 @@ sub strand {
 
 =head2 frame
 
-See documentation in L<Bio::Search::SearchUtils::frame()|Bio::Search::SearchUtils>
+See documentation in L<Bio::Search::Hit::HitI::frame()|Bio::Search::Hit::HitI>
 
 =cut
 
@@ -1377,10 +1485,16 @@ sub frame {
 #----------
     my( $self, $frm ) = @_;
 
-    Bio::Search::SearchUtils::tile_hsps($self) if not $self->{'_tiled_hsps'};
+    unless ($self->{'_hsps'}) {
+        Bio::Search::SearchUtils::_warn_about_no_hsps($self);
+        #return wantarray ? ('-','-') : '-';
+        return '-';
+    }
+
+    Bio::Search::SearchUtils::tile_hsps($self) unless $self->tiled_hsps;
 
     if( defined $frm ) {
-	return $self->{'_frame'} = $frm;
+        return $self->{'_frame'} = $frm;
     }
 
     # The check for $self->{'_frame'} is a remnant from the 'query' mode days
@@ -1388,9 +1502,9 @@ sub frame {
 
     my ($frame);
     if(not defined($self->{'_frame'})) {
-	$frame = $self->hsp->frame;
+        $frame = $self->hsp->frame;
     } else {
-	$frame = $self->{'_frame'}; 
+        $frame = $self->{'_frame'}; 
     } 
     return $frame;
 }
@@ -1429,14 +1543,14 @@ sub locus {
     if( defined $value || ! defined $previous ) { 
       unless (defined $value) {
         if ($self->{'_name'} =~/(gb|emb|dbj|ref)\|(.*)\|(.*)/) {
-		  $value = $previous = $3;
-		} else {
+                  $value = $previous = $3;
+                } else {
           $value = $previous = '';
         }
       }
-	  $self->{'_locus'} = $value;
+          $self->{'_locus'} = $value;
     } 
-	return $previous;
+        return $previous;
 }
 
 =head2 each_accession_number
@@ -1461,13 +1575,17 @@ sub each_accession_number {
       while ($desc =~ /(\b\S+\|\S*\|\S*\s?)/g) {
         my $id = $1;
         my ($acc, $version);
-	    if ($id =~ /(gb|emb|dbj|sp|pdb|bbs|ref|lcl)\|(.*)\|(.*)/) {
-	      ($acc, $version) = split /\./, $2; 
-	    } elsif ($id =~ /(pir|prf|pat|gnl)\|(.*)\|(.*)/) {
-	      ($acc, $version) = split /\./, $3;  
-	    } else {
-	   	  #punt, not matching the db's at ftp://ftp.ncbi.nih.gov/blast/db/README
-	   	  #Database Name                     Identifier Syntax
+	if ($id =~ /(gb|emb|dbj|sp|pdb|bbs|ref|tp[gde])\|(.*)\|(.*)/) {
+	    ($acc, $version) = split /\./, $2; 
+	} elsif ($id =~ /(pir|prf|pat|gnl)\|(.*)\|(.*)/) {
+	    ($acc, $version) = split /\./, $3;  
+	} elsif( $id =~ /(gim|gi|bbm|bbs|lcl)\|(\d*)/) {
+	    $acc = $id;
+	} elsif( $id =~ /(oth)\|(.*)\|(.*)\|(.*)/ ) { # discontinued...
+	    ($acc,$version) = ($2);
+	} else {
+                     #punt, not matching the db's at ftp://ftp.ncbi.nih.gov/blast/db/README
+                     #Database Name                     Identifier Syntax
           #============================      ========================
           #GenBank                           gb|accession|locus
           #EMBL Data Library                 emb|accession|locus
@@ -1478,20 +1596,20 @@ sub each_accession_number {
           #Brookhaven Protein Data Bank      pdb|entry|chain
           #Patents                           pat|country|number 
           #GenInfo Backbone Id               bbs|number 
-          #General database identifier	   gnl|database|identifier
+          #General database identifier           gnl|database|identifier
           #NCBI Reference Sequence           ref|accession|locus
           #Local Sequence identifier         lcl|identifier
-	      $acc=$id;
-	    }
-	    push(@accnums, $acc);
-	  }
+              $acc=$id;
+            }
+            push(@accnums, $acc);
+          }
     }  
     return @accnums;
 }
 
 =head2 tiled_hsps
 
-See documentation in L<Bio::Search::SearchUtils::tiled_hsps()|Bio::Search::SearchUtils>
+See documentation in L<Bio::Search::SearchUtils::tile_hsps()|Bio::Search::SearchUtils>
 
 =cut
 
@@ -1501,4 +1619,23 @@ sub tiled_hsps {
     return $self->{'_tiled_hsps'};
 }
 
+=head2 query_length
+
+ Title   : query_length
+ Usage   : $obj->query_length($newval)
+ Function: Get/Set the query_length
+ Returns : value of query_length (a scalar)
+ Args    : on set, new value (a scalar or undef, optional)
+
+
+=cut
+
+sub query_length{
+    my $self = shift;
+
+    return $self->{'_query_length'} = shift if @_;
+    return $self->{'_query_length'};
+}
+
+
 1;
